/* 
  Copyright (C) 1998 Norbert Preining and Christofer Edvardsen

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "smmgr.h"

int ulfile(const char *s) {
	int infile;
	int n, m;
	char buf[256], endstr[] = { 0x10, 3 };
	
	if ( ( infile = open(s,O_RDONLY ) ) < 0 ) {
		logit(L_ERROR,"Couldn't open file %s.\n",s);
		return(1);
	}

	while ( ( n = read (infile,buf, sizeof(buf) )) > 0) {
		for ( m = 0; m<n; m++) {
			write (fd, &buf[m], 1);
			if ( buf[m] == 0x10 ) write (fd, &buf[m], 1);
		}
	}
	write(fd, endstr, sizeof(endstr));
	return(0);
} /* ulfile */


int report_msgs() 
{
	int i;
	for (i=0;i<nr_msgs;i++) {
		printf("Message Index %d:\n",i+1);
		printf("\tType:      %6d (%4x)\n",
		       msginfos[i].type,msginfos[i].type);
		printf("\tPage/Time: %6d (%4x)\n",
		       msginfos[i].pagetime,msginfos[i].pagetime);
		printf("\tAttribut:  %6d (%4x) - ",
		       msginfos[i].attr,msginfos[i].attr);
		switch(msginfos[i].attr) {
		case 255: printf("unreleased, unchecked\n");
			break;
		case 253: printf("unreleased, checked\n");
			break;
		case 252: printf("released, checked\n");
			break;
		default:  printf("unknown\n");
		}
		printf("\tStatus:    %6d (%4x)\n",
		       msginfos[i].status,msginfos[i].status);
		printf("\tDay:       %6d (%4x)\n",
		       msginfos[i].day,msginfos[i].day);
		printf("\tHour:      %6d (%4x)\n",
		       msginfos[i].hour,msginfos[i].hour);
		printf("\tMinute:    %6d (%4x)\n",
		       msginfos[i].min,msginfos[i].min);
		printf("\tStartpage: %6d (%4x)\n",
		       msginfos[i].startpage,msginfos[i].startpage);
		printf("\tAddress:   %6d (%4x)\n",
		       msginfos[i].address,msginfos[i].address);
		printf("\tChecksum:  %6d (%4x)\n",
		       msginfos[i].checksum,msginfos[i].checksum);
		printf("\tID:        \"%s\"\n",msginfos[i].id);
	}
	return(0);
}

int listmsg() {
	int n;
	char type[100];
	char rtime_str[256], rtime_short[32];

	for (n=0;n<nr_msgs;n++) {
		make_rtime(msginfos[n].day, msginfos[n].hour, 
			   msginfos[n].min,
			   &rtime_short[0], &rtime_str[0]);
		attr2str(msginfos[n].attr,type);
		switch(msginfos[n].type) {
		case 1:
			printf("%2d: Fax;    Pages: %2dp; Time:%s; Status: %s\n",
			       n+1,msginfos[n].pagetime,rtime_short,type);
			break;
		case 2: 
			printf("%2d: Voice; Length: %2ds; Time:%s; Status: %s\n",
			       n+1,msginfos[n].pagetime,rtime_short,type);
			break;
		default:
			printf("%2d: Unrecognized message format\n",n);
		}
	}
	return(0);
}


/*
  Get first free message number
*/
int freenr() {
	DIR *dp;
	struct dirent *ep;
	int n=1, m;
	char strtmp[256];
     
	strcpy(strtmp,s_faxprefix);
	strcat(strtmp,"%dp");
	dp = opendir (s_faxdir);
	if (dp != NULL)
	{
		while ((ep = readdir (dp))) {
			m = 0;
			sscanf (ep->d_name,strtmp,&m);
			if (m>=n) n=m+1;
		}
		(void) closedir (dp);
	} else {
		logit(L_ERROR,"Couldn't open directory %s.",s_faxdir);
		exit(1);
	}
	dp = opendir (s_voicedir);
	if (dp != NULL)
	{
		while ((ep = readdir (dp))) {
			m=0;
			sscanf (ep->d_name,"voice%d.",&m);
			if (m>=n) n=m+1;
		}
		(void) closedir (dp);
	} else {
		logit(L_ERROR,"Couldn't open directory %s.",s_voicedir);
		exit(1);
	}
	return (c_firstfree=n);
} /* freenr */


/*
  Read all messages from modem and write them to files
*/

int msgtofile() {
	int i;
	int ret = 0;
	for (i=0;i < nr_voice + nr_fax; i++) 
		ret |= onetofile(i+1);
	return(ret);
}


/*
  savepagetofile (reines debug ding!)
*/

int savepagetofile(int n, const char *fn) {
	FILE *msgfile;

	get_page(n);
	msgfile = fopen(fn,"w");
	if ( msgfile == NULL ) {
		logit(L_ERROR,"Failed to open messagefile for output: %s\n",fn); 
		return(-1);
	} else {
		logit(L_DEBUG,"Writing page %d to %s.\n",n,fn);
	}
	fwrite(&page[0],1,32772,msgfile);
	fclose(msgfile);
	return(0);
} /* savepagetofile */

/*
  msgnewtofile
*/
int newtofile() {
	int i;
	int ret=0;
	
	for (i=0;i < nr_voice + nr_fax; i++) 
		if (is_new_or_unknown(i)) ret |= onetofile(i+1);
	return(ret);
}


/*
  newvoicetofile
*/
int newvoicetofile() {
	int i;
	int ret=0;
	
	for (i=0;i < nr_voice + nr_fax; i++) 
		if (is_new_or_unknown(i) && is_voice(i)) ret |= onetofile(i+1);
	return(ret);
}

/*
  newfaxtofile
*/
int newfaxtofile() {
	int i;
	int ret=0;
	
	for (i=0;i < nr_voice + nr_fax; i++) 
		if (is_new_or_unknown(i) && is_fax(i)) ret |= onetofile(i+1);
	return(ret);
}

/* 
   onetofile
*/
int onetofile(int nn) {
	int c_index, c_msg, c_type, c_length, n_page, n_offset;
	time_t rtime_day, rtime_hour, rtime_min;
	time_t file_day, file_hour,file_min;
	int file_length, file_type;
	char msgname[256], commentname[256], title[256];
	char rtime_str[256], rtime_short[32];
	int faxpage, start_page, start_offset;
	int ret=0;
	
	start_page = msginfos[nn-1].startpage;
	start_offset = msginfos[nn-1].address;
	c_type = msginfos[nn-1].type;
	
	c_offset = 0;

	ret |= get_page(start_page);

	c_msg = nn;
	/* shift offset to beginning +2 (2 bytes checksum) */
	/* stupid me: already killed in get_page!!!!!!!!! */
	/* c_index=incoffset( start_offset + 2 ); */
	c_index=incoffset( start_offset );
	/* The message has the following format: */
	/* Start Length
	   0	1	Message Index
	   1	1	Message Type
	   2	1	Message Information
	   3	1	Message Attributes
	   4	1	Reception Status
	   5	1	Time Stamp: Days
	   6	1	Time Stamp: Hours
	   7	1	Time Stamp: Minutes
	   8	20	Sender Fax ID
	   28	1	Previous Message: Page Number
	   29	1	Previous Message: Address
	   30	1	Previous Message: Address
	   31	1	Next Message: Page Number
	   32	1	Next Message: Address
	   33	1	Next Message: Address
	   34	n	Message
	*/
	/* incoffset( 1 ); */               file_type = incoffset( 1 );
	c_length = msginfos[nn-1].pagetime; file_length=incoffset( 1 ); 
	rtime_day = msginfos[nn-1].day;     file_day=incoffset( 3 ); 
	rtime_hour = msginfos[nn-1].hour;   file_hour=incoffset( 1 );
	rtime_min = msginfos[nn-1].min;     file_min=incoffset( 1 );
	/*	incoffset(6);*/
	incoffset(20);
	n_page=incoffset( 4 );
	n_offset=incoffset( 1 )*256;
	n_offset+=incoffset( 1 );
	logit(L_DEBUG,"onetofile: computed values: c_index=%d,n_page=%d,n_offset=%d,c_length=%d,file_length=%d\n",c_index,n_page,n_offset,c_length,file_length);
	incoffset( 1 );

	faxpage=1;

	ret |= make_rtime(rtime_day, rtime_hour, rtime_min,
			  &rtime_short[0], &rtime_str[0]);
	ret |= make_names(&msgname[0],&commentname[0],&title[0],
			  c_msg,rtime_short);
	ret |= make_commentfile(commentname, c_msg, 
				rtime_short, rtime_str, title);

	strcpy(current_messagename,msgname);
	if (s_ignore_dle) {
		ret |= transfer_no_dle_handle(n_page, n_offset, msgname, c_msg, commentname);
	} else {
		switch (c_type) {
		case T_VOICE:
			ret |= transfer_voice(n_page, n_offset, msgname, c_msg, commentname);
			break;
		case T_FAX:
		default:
			ret |= transfer_fax(n_page, n_offset, msgname, c_msg, commentname);
			break;
		}
	}
	msginfos[nn-1].attr &= ATTR_RELEASED;
	c_firstfree++;
	return(ret);
}

/*
	int view_fax(int l)

	view a fax
*/
int view_fax(int index) {
	char save_faxdir[256];
	char fname[256];
	struct stat ds;
	int nextfree=c_firstfree==-1?freenr():c_firstfree;
	int ret;
	
	strcpy(save_faxdir,s_faxdir);
	strcpy(s_faxdir,tmpdir);
	if (stat(s_faxdir,&ds)!=0) {
		if (errno == ENOENT) {
			if (mkdir(s_faxdir,0755)!=0) {
				logit(L_ERROR,"Cannot mkdir %s!\n",s_faxdir);
				strcpy(s_faxdir,save_faxdir);
				return(1);
			} else {
				stat(s_faxdir,&ds);
			}
			
		} else {
			logit(L_ERROR,"Cannot stat in directory /tmp!? - ERRNO: %d\n",errno);
			strcpy(s_faxdir,save_faxdir);
			return(1);
		}
	}
	if (!S_ISDIR(ds.st_mode)) {
		logit(L_ERROR,"%s is not a directory - why?\n",s_faxdir);
		strcpy(s_faxdir,save_faxdir);
		return(1);
	}
	if (index2fnr[index] == -1) {
		if (onetofile(index)==0) {
			index2fnr[index] = nextfree;
		} else {		
			strcpy(s_faxdir,save_faxdir);
			return(1);
		}
	}
	snprintf(fname,256,"%s/%s%04d",s_faxdir,s_faxprefix,index2fnr[index]);
	ret = viewit(fname,index);
	strcpy(s_faxdir,save_faxdir);
	return(ret);
}
		
/*
  int viewit(char *)

  effectively view file
*/
int viewit(const char* fname,int index) {
        int i,kl,len;
        char sys[512],tmpstr[20];
        char *p;
	char **fnamelist;

/* viewprog string:
   %f       ->     datafile1 datafile2 ...
   %(str)F   ->     str datafile1 str datafile2 ...
*/
        p = s_viewprog;
        strcpy(sys,"");
        strcpy(tmpstr,"");

/* generate fnamelist */
	len = msginfos[index].pagetime;
	fnamelist = (char**)malloc(len*sizeof(char*));
	for (i=1;i<=len;i++) {
		fnamelist[i-1] = (char*)malloc(256*sizeof(char));
		snprintf(fnamelist[i-1],256,"%sp%03d.g3",fname,i);
	}
/* build system string */
        kl = 0;
        while (*p != '\0') {
                if (*p != '%') {
                        strncat(sys,p,1);
                } else {
                        p++;
                        switch (*p) {
                        case 'f': for (i=0; i<len; i++) {
                                strcat(sys,fnamelist[i]);
                                strcat(sys," ");
                        }
                        break;
                        case '(': p++;
                                while (*p != ')' || kl > 0) {
                                        strncat(tmpstr,p,1);
                                        if (*p == '(') kl++;
                                        if (*p == ')') kl--;
                                        p++;
                                }
                                p++;
                                if (*p != 'F') {
                                        logit(L_ERROR,"Error Format: %%(str)F\n");
                                        return(-1);
                                }
                                for (i=0; i<len; i++) {
                                        strcat(sys,tmpstr);
                                        strcat(sys,fnamelist[i]);
                                        strcat(sys," ");
                                }
                                break;
                        default: logit(L_ERROR,"Error Format: %%f, %%(str)F recognized!\n");
                                return(-1);
                                break;
                        }
                }
                p++;
        }
        for (i=0; i<len; i++) free(fnamelist[i]);
        free(fnamelist);
        return(system(sys));
}



#ifdef USEGSM

/*
  Decodes len bytes into buffer buf
*/
int usr2gsm(char *buf, int len) {
	int i=0;
	int prev_dle=0;
	unsigned char c;
	c = incoffset(0);
	while (i<len) {
		if ( prev_dle == 1) {
			switch(c) {
			case 0x10:
				bcopy(&page[c_offset],&(buf[i++]),1);
				break;
			case 0x18:
			case 0x03:
				logit(L_DEBUG,"transfer: Got <DLE><CAN> or <DLE><ETX>, looks like end ;-)\n");
				return(0);
				break;
			default:
				logit(L_ERROR,"After <DLE>: 0x%x\n",c);
				logit(L_DEBUG,"page=%d, offset=%d\n",c_page,c_offset);
			}
			prev_dle = 0;
		} else {
			if ( c == 0x10 ) {
				prev_dle = 1;
			} else {
				bcopy(&c,&(buf[i++]),1);
			}
		}
		c = incoffset(1);
	}
	return(i);
}
			

/*
  playtodsp
  plays a voice-msg direct to dsp, without saving it first
  needs libgsm
*/
int play_soundcard(int index) {
       gsm handle;
       gsm_frame buf;
       gsm_signal sample[160];
       int cc;
       int i;
       FILE *soundf, *debugf;
       int start_page;
       int start_offset;
       int n_page, n_offset;

       if (index == 0) {
	       start_page = 0;
	       start_offset = 0;
	       n_page = 0;
	       n_offset = 32768;
       } else if (index == 255) {
	       start_page = 1;
	       start_offset = 0;
	       n_page = 1;
	       n_offset = 32512;
       } else {
	       start_page = msginfos[index-1].startpage;
	       start_offset= msginfos[index-1].address;
       }
       get_page(start_page);
       c_offset = 0;
       incoffset(start_offset);
       if (index != 0 && index != 255) {
	       n_page = incoffset(31);
	       for (i=start_page+1; i<=n_page; i++) {
		       get_page(i);
	       }
	       get_page(start_page);
	       c_offset = 0;
	       incoffset(start_offset+31);
	       n_offset  = incoffset(1)*256;
	       n_offset += incoffset(1);
	       incoffset(1); /* ???? WHY, but it is in transfer voice!!! */
       }
       /* now we SHOULD be at the gsm-data! */
       if (!(handle = gsm_create())) {
	       logit(L_ERROR,"Cannot create GSM-handle, cannot play directly!\n");
	       return(1);
       }
	if (s_verbose == 3) {
		debugf = fopen("debug.ul","w");
	}
       /* now open sound fd */
       if ((soundf = fopen("/dev/audio","w")) == NULL) {
	       logit(L_ERROR, "Cannot open /dev/audio: %s\n",strerror(errno));
	       gsm_destroy(handle);
	       return(1);
       }
       /* we compare c_offset+2 < n_offset, such that the last 2 bytes 0x10 0x18
	  do NOT start a new frame! */
       while (((c_page < n_page) || (c_offset+2 < n_offset)) && ((cc = usr2gsm((char *)buf, sizeof buf)) > 0)) {
	       logit(L_DEBUG,"play_soundcard: c_page=%d, n_page=%d, c_offset=%d, n_offset=%d, cc=%d\n",c_page, n_page, c_offset, n_offset,cc);
	       if (cc != sizeof buf) {
		       logit(L_ERROR,"gsm-decode: invalid buffer size!\n");
		       gsm_destroy(handle);
		       return(1);
	       }
	       if (gsm_decode(handle, buf, sample) < 0) {
		       logit(L_ERROR,"Cannot decode gsm-sample!\n");
	       }
	       /* now do a ulaw output, copied from toast/gsm-lib */
	       ulaw_output(sample,soundf);
		if (s_verbose == 3) {
			ulaw_output(sample,debugf);
		}
       }
	if (s_verbose == 3) {
		fclose(debugf);
	}
       fclose(soundf);
       gsm_destroy(handle);
       return(0);
}

#else /* NOW WITHOUT GSM-LIB */

/*
	int play_soundcard (int index)
*/

int play_soundcard (int index) {
	char save_voicedir[256];
	char fname[256];
	struct stat ds;
	int nextfree=c_firstfree==-1?freenr():c_firstfree;
	int ret;
	
	strcpy(save_voicedir,s_voicedir);
	strcpy(s_voicedir,tmpdir);
	if (stat(s_voicedir,&ds)!=0) {
		if (errno == ENOENT) {
			if (mkdir(s_voicedir,0755)!=0) {
				logit(L_ERROR,"Cannot mkdir %s!\n",s_voicedir);
				strcpy(s_voicedir,save_voicedir);
				return(1);
			} else {
				stat(s_voicedir,&ds);
			}
			
		} else {
			logit(L_ERROR,"Cannot stat in directory /tmp!? - ERRNO: %d\n",errno);
			strcpy(s_voicedir,save_voicedir);
			return(1);
		}
	}
	if (!S_ISDIR(ds.st_mode)) {
		logit(L_ERROR,"%s is not a directory - why?\n",s_voicedir);
		strcpy(s_voicedir,save_voicedir);
		return(1);
	}
	if (index2fnr[index] == -1) {
		if (onetofile(index)==0) {
			index2fnr[index] = nextfree;
		} else {		
			strcpy(s_voicedir,save_voicedir);
			return(1);
		}
	}
	snprintf(fname,256,"%s/voice%04d.gsm",s_voicedir,index2fnr[index]);
	ret = playit(fname);
	strcpy(s_voicedir,save_voicedir);
	return(ret);
}

/*
  int playit(char *)

  effectively play file
*/
int playit(const char* fname) {
	char s[500];
	snprintf(s,500,"play -t gsm %s &",fname);
	return(system(s));
}

#endif
