/* 
  Copyright (C) 1998 Norbert Preining and Christofer Edvardsen

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "smmgr.h"

/*
  Open serial device and configure it
*/
int opendev() {
	char devicename[32], tmpstr[32];
	int lockfile;

	sprintf(lockname,"%s/LCK..%s",s_lockdir,s_device);

	if ( ( lockfile=open(lockname,O_RDONLY ) ) >= 0 ) {
		logit(L_ERROR,"Device %s in use.\n",s_device);
		logit(L_ERROR,"Lockfile: %s\n",lockname);
		lockname[0]=0;
		close(lockfile);
		exit(0);
	}

	if ( ( lockfile=open(lockname,O_WRONLY | O_CREAT | O_EXCL, 0664 ) ) < 0 ) {
		lockname[0]=0;
                logit(L_ERROR,"Couldn't create lockfile: %s\n",lockname);
                exit(1);
        }
	
	sprintf(tmpstr,"%10d smmgr %s\n", getpid(),getlogin());
	write(lockfile,tmpstr,strlen(tmpstr));

	close(lockfile);

	sprintf(devicename,"/dev/%s",s_device);
	fd = open(devicename, O_RDWR | O_NOCTTY | O_NONBLOCK );
	if ( fd < 0 ) {
		logit(L_ERROR,"Couldn't open device.\n");
		exit (1);
	}

	tcgetattr(fd,&oldiosettings);
	tcgetattr(fd,&iosettings);

	iosettings.c_iflag = 0;
	iosettings.c_oflag = 0;

	iosettings.c_cflag = CRTSCTS | CS8 | CLOCAL | CREAD;
	iosettings.c_lflag = 0;

	iosettings.c_cc[VTIME] = 1;
	iosettings.c_cc[VMIN] = 0;

	if (cfsetispeed(&iosettings,s_baud)==-1) {
		logit(L_ERROR,"Error setting port ispeed.\n");
		exit(1);
	} else logit(L_DEBUG,"Port ispeed set to %d.\n",s_baud);
	if (cfsetospeed(&iosettings,s_baud)==-1) {
		logit(L_ERROR,"Error setting port ospeed.\n");
		exit(1);
	} else logit(L_DEBUG,"Port ospeed set to %d.\n",s_baud);

	tcflush(fd, TCIOFLUSH);
	tcsetattr(fd, TCSANOW, &iosettings);
	
	return(0);
} /* opendev */

int closedev() {
	if (fd>=0) {
		sendstring("AT",NULL,NULL);
		tcsetattr(fd, TCSANOW, &oldiosettings);
		tcflush(fd, TCIOFLUSH);
		close(fd);
	}	
	if ((lockname[0]!=0)&&(unlink(lockname)==-1)) {
		logit(L_ERROR,"Couldn't remove lockfile: %s\n",lockname);
		return(1);
	}	
	return(0);
}


/*
  Send a string to modem, read line by line, wait for \nOK or \nERROR, 
  and return the string the modem answered.
*/
modemval sendstring(char *sendstr, char **retret, char *fine) {
	char tmpbuf[256],c;
	int spaceleft, totalsize, tl, tmpbuf_end, finished = 0, got, got_newline, n, nt;
	modemval retval = M_NORESPONSE;
	int in_ok = 0, in_error = 0, nr_error = 0, nr_ok = 0, keepon = 500;
	int in_fine = 0, nr_fine = 0, fine_length;
	char okstr[] = "OK", errorstr[] = "ERROR";
	char *retbuf;
	int initsize=40;

	if (fine != NULL) {
		fine_length = strlen(fine);
	}

	/* initialize the return buffer to a value which should
	   fit most return answers */
	if (retret != NULL) {
		retbuf = (*retret) = (char*) malloc(initsize*sizeof(char));
		retbuf[0] = '\0';
		spaceleft = initsize-1;
		totalsize = initsize;
	}

	/* send sendstr */
	if ( sendstr == NULL ) {
		logit(L_DEBUG,"Sending newline.\n");
	} else {
		logit(L_DEBUG,"Sending \"%s\".\n",sendstr);
		write(fd,sendstr,strlen(sendstr));
	}
	write(fd,"\n\r",2);

	got_newline = 0;
	/* start read loop */
	while ( finished == 0 && keepon >= 0) {
		got = read(fd, tmpbuf, sizeof(tmpbuf));
		if ( got > 0 ) {
			tmpbuf_end = got;
			for (n=0; n < got; n++) {
				c = tmpbuf[n];
				if (isascii(c)) {
					logit(L_DEBUG,"==>%c\n",c);
				}
				if (got_newline == 1) {
					if (c == 'O') {
						in_ok = 1;
						nr_ok = 1;
					}
					if (c == 'E') {
						in_error = 1;
						nr_error = 1;
					}
					if (fine != NULL && c == fine[0]) {
						in_fine = 1;
						nr_fine = 1;
					}
					if (c != 'O' && c != 'E' && fine != NULL && c != fine[0]) {
							in_ok = 0;
							in_error = 0;
							in_fine = 0;
					}
					if (c == '\n') got_newline = 1;
					else got_newline = 0;
				} else {
					if (in_ok) {
						if (nr_ok <2 && c == okstr[nr_ok]) {
							if (nr_ok == 1) {
								finished = 1;
								retval = M_OK;
								tmpbuf_end = n+1;
								n = got; /* break for loop */
							}
						} else {
							in_ok = 0;
						}
					}
					if (in_error) {
						if (nr_error < 5 && c == errorstr[nr_error]) {
							nr_error++;
							if (nr_error == 5) {
								finished = 1;
								retval = M_ERROR;
								tmpbuf_end = n+1;
								n = got; /* break for loop */
							}
						} else {
							in_error = 0;
						}
					}
					if (in_fine) {
						if (nr_fine < fine_length && c == fine[nr_fine]) {
							nr_fine++;
							if (nr_fine == fine_length) {
								finished = 1;
								retval = M_OK;
								tmpbuf_end = n+1;
								n = got;
							}
						} else {
							in_fine = 0;
						}
					}
					if (tmpbuf[n] == '\n') got_newline = 1;
					else got_newline = 0;
				}
			}
			/* else copy return string to return buffer */
			if ((tl = tmpbuf_end)>0 && retret != NULL) {
				if (tl <= spaceleft) {
					strncat(retbuf,tmpbuf,tl);
					spaceleft -= tl;
				} else {
					logit (L_INFO,"Realloc retbuf: old size=%d\n", totalsize); 
					nt = 2*(tl>totalsize?tl:totalsize);
					retbuf = (*retret) = (char*) realloc (*retret, nt);
					if (retbuf == (char*)NULL) {
						logit(L_ERROR, "Cannot reallocate memory!\n");
						exit(1);
					}
					spaceleft += nt - totalsize;
					totalsize = nt;
					strncat(retbuf,tmpbuf,tl);
					spaceleft -= tl;
				}
			}
			keepon = 500;
		} else {
			keepon--;
			usleep(2000);
		}
	}
	if (retret != NULL && retval != M_NORESPONSE) {
		/* Now kill leading and final junk, including OK/ERROR */
		tl = strlen(retbuf);
		got = 0; /* the dest index */
		n = 0;
		while ((retbuf[n] == '\n' || retbuf[n] == '\r' || retbuf[n]=='\xca' || retbuf[n]=='\x8a' || retbuf[n]=='\xfe' || retbuf[n]=='\xea' || retbuf[n]=='\xfa' || retbuf[n]=='\xff')) n++;
		/* n is now the startingposition of the real retval or 
		 * (if ATE1) on the starting of the sendstring */
		/* Check if sendstring is there */
		if (strncmp(sendstr, &retbuf[n], strlen(sendstr)) == 0) {
			n += strlen(sendstr);
			/* Skip over junk after sendstr */
			while ((retbuf[n] == '\n' || retbuf[n] == '\r' || retbuf[n]=='\xca' || retbuf[n]=='\x8a' || retbuf[n]=='\xfe' || retbuf[n]=='\xea' || retbuf[n]=='\xfa' || retbuf[n]=='\xff')) n++;
		}
		if (retval == M_OK) {
			if (fine != NULL) got = tl - (strlen(fine)+2); 
			else got = tl - (2+2);
		} else {
			got = tl - (5+2);
		}
		for (tl = 0; tl < got - n; tl++) retbuf[tl] = retbuf[tl+n];
		retbuf[tl++] = '\0';
	}
	return (retval);
}


/* 
   Get a block binary data with a specified size from modem
*/

char *getblock(char *sendstr, int expect)
{
	int res=1, n=0, keepon;
#ifdef HELMA
	int sendstrlen;
#endif
	char *retbuf;

	retbuf=(char*) malloc(expect);
	if (retbuf==NULL) {
		logit(L_ERROR,"Out of memory.\n");
		exit(1);
	}

	/* Make sure no characters are waiting in the buffer */
	keepon = 5;
	while ( keepon > 0 ) {
		res = read( fd, retbuf, expect );    
		if ( res > 0 ) {
			keepon= 5;
		} else {
			keepon--;
			usleep(5000);
		}
	}
  
	logit(L_DEBUG,"Sending \"%s\".\n",sendstr);
#ifdef HELMA
	sendstrlen = strlen(sendstr);
#endif
	write(fd,sendstr,strlen(sendstr));
	write(fd,"\n\r",2);
#ifdef HELMA
	sendstrlen += 2;

	/* On the Professional Message Modem we have 'sendstr\n\r'
	   at the beginning of the block. Let's get it. */
#endif
	keepon = 1000;
#ifdef HELMA
	while ( keepon > 0 ) {
		res = read( fd, &retbuf[n], sendstrlen-n );
		if ( res > 0 ) {
			n += res;
			keepon = 50;
		} else {
			logit(L_DEBUG,":");
			keepon--;
			usleep(2000);
		}
		if (n>sendstrlen) {
			logit(L_ERROR,"Got too many characters. Shouldn't happen.\n");
			free(retbuf);
			return NULL;
		}
		if (n==sendstrlen) keepon=0;
	}
	/* If we have 'sendstr' at the beginning of the block,
	   discard it. */
	if (strncmp(retbuf, sendstr, sendstrlen-2) == 0) {
		logit(L_DEBUG,"Discard command \"%s\" from block.\n", sendstr);
		n = 0;
	}
	keepon = 1000;
#endif
	while ( keepon > 0 ) {
		res = read( fd, &retbuf[n], expect-n );    
		if ( res > 0 ) {
			logit(L_DEBUG,"\nGot %5d.\n",res);
			n += res;
			keepon = 50;
		} else {
			logit(L_DEBUG,".");
			keepon--;
			usleep(2000);
		}
		if (n>expect) {
			logit(L_ERROR,"Got too many characters. Shouldn't happen.\n");
			free(retbuf);
			return NULL;
		}
		if (n==expect) keepon=0;
	}
	getblock_return = n;
	if ( n != expect ) {
		logit(L_ERROR,"Only got %d bytes (wanted %d).\n",n,expect);
		return retbuf;
	}
	logit(L_DEBUG,"Successfully got %d bytes.\n",n);
	return retbuf;
} /* getblock */

int getmodemstatus() {
	char *ret=NULL;
	char *p;
	int n,i;
	
	logit(L_INFO,"Getting status from modem.\n");
	/* MCS */
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCS?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcs = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcs = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCD?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcd = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcd = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCF?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcf = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcf = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCL?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcl = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcl = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCM?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcm = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcm = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCP?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcp = str2num(ret,4); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcp = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCR?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcr = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcr = -1;
	}
	for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCV?",&ret,NULL);
	if (ret != NULL) {
		modemstat.mcv = str2num(ret,1); 
		free(ret); ret = NULL;
	} else {
		modemstat.mcv = -1;
	}
	if (modemtypes[s_modemtype].id == USR_MESSAGE_PRO) {
		for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCT?",&ret,NULL);
		if (ret != NULL) {
			modemstat.mct = str2num(ret,1); 
			free(ret); ret = NULL;
		} else {
			modemstat.mct = -1;
		}
		for (n=0; (n<4) && (ret==NULL); n++) sendstring("AT+MCG?",&ret,NULL);
		if (ret != NULL) {
			p = ret; i = 0;
			while (*p >= '0' && *p <= '9') { p++; i++; };
			modemstat.mcg = str2num(ret,i); 
			free(ret); ret = NULL;
		} else {
			modemstat.mcg = -1;
		}
	}
	for (n=0; (n<4) && (ret==NULL); n++) 
		sendstring("AT+MVD?",&ret,NULL);
	p = ret; i=0;
	while (*p >= '0' && *p <= '9') { p++; i++; };
	modemstat.mvd = str2num(ret,i); free(ret); ret = NULL;

	for (n=0; (n<4) && (ret==NULL); n++) 
		sendstring("AT+MCC?",&ret,NULL);
	strncpy(modemstat.mcc,ret,15); free(ret); ret = NULL;
	for (n=0; (n<4) && (ret==NULL); n++) 
		sendstring("AT+MFI?",&ret,NULL);
	strcpy(modemstat.mfi,ret); free(ret); ret = NULL;
	return(0);
}

int getmsginfo() {
	char *msginfo=NULL;
	int n, addr_high_byte, addr_low_byte; 
	char tmpstr[80];
	
	logit(L_INFO,"Getting message info from modem.\n");
	for (n=0; (n<4) && (msginfo==NULL); n++) {
		sendstring("AT+MSR=0",&msginfo,NULL);
	}
	total_memory = str2num(&msginfo[0],3);
	percentfull  = str2num(&msginfo[4],3);
	nr_voice     = str2num(&msginfo[8],3);
	nr_voice_new = str2num(&msginfo[12],3);
	nr_fax       = str2num(&msginfo[16],3);
	nr_fax_new   = str2num(&msginfo[20],3);
	inmsg        = str2num(&msginfo[24],3);
	fullmsg      = str2num(&msginfo[28],3);
	free(msginfo);
	msginfo=NULL;

	nr_msgs = nr_voice + nr_fax;

	msginfos = malloc(nr_msgs*sizeof(struct msginfo_));
	
	for (n=0;n<nr_msgs;n++) {
		while (msginfo==NULL) {
			sprintf(tmpstr,"AT+MSR=%d",n+1);
			sendstring(tmpstr,&msginfo,NULL);
		}
		msginfos[n].type           = str2num(&msginfo[4],3);
		msginfos[n].pagetime       = str2num(&msginfo[8],3);
		msginfos[n].attr           = str2num(&msginfo[12],3);
		msginfos[n].status         = str2num(&msginfo[16],3);
		msginfos[n].day            = str2num(&msginfo[20],3);
		msginfos[n].hour           = str2num(&msginfo[24],3);
		msginfos[n].min            = str2num(&msginfo[28],3);
		strncpy(msginfos[n].id,&msginfo[32],20);
		msginfos[n].id[20]=0;
		msginfos[n].startpage      = str2num(&msginfo[53],3);
		addr_high_byte             = str2num(&msginfo[57],3);
		addr_low_byte              = str2num(&msginfo[61],3);
		msginfos[n].address = addr_high_byte*256+addr_low_byte;
		msginfos[n].checksum       = str2num(&msginfo[65],3);

		free(msginfo);
		msginfo=NULL;
	}
	return(0);
} /* getmsginfo */

unsigned char incoffset( int toadd )
{
	c_offset+=toadd;
	if ( c_offset >= pages[c_page].length ) {
		c_offset -= pages[c_page].length;
		c_page++;
		/* Professional Message Modem seems to be a bit fuzzy.
		 * If I don't wait, we will receive nothing */
/* 		if (modemtypes[s_modemtype].id == USR_MESSAGE_PRO) */
/* 		usleep(50000); */
		get_page(c_page);
	}
	return (pages[c_page].data)[c_offset];
}

/*
  Reset the internal clock of the modem and store the current time
  to a file to allow translations from the modem time to real world time
  The time stored in the libfile ist 00:00 of the current day.
*/
int resetclock(int writelib)
{
	time_t hour, min, sec, day, difftime;
	char tmpstr[64];

	if (writelib == 1) {
		day = time(NULL);
		hour = ( day % ( 3600*24 ) ) / 3600 ;
		min = ( day % 3600 ) / 60 ;
		sec = day % 60;
		day -= hour*3600 + min*60 + sec;
		set_stored_time(day);
	} 
	get_stored_time();
	difftime = time(NULL) - stored_time;
	day = difftime / (3600*24);
	hour = (difftime % ( 3600*24 ) ) / 3600;
	min = ( difftime % 3600 ) / 60 ;
	sec = difftime % 60;
	sprintf(tmpstr,"AT+MCC=%d,%d,%d,%d", (int)day,(int)hour,
					     (int)min,(int)sec);
	sendstring(tmpstr, NULL,NULL);
	return(0);
}

/*
  Read modem time and determine if it is a sane value
*/
int readclock()
{
	time_t day, hour, min, sec;
	char *modemtime;

	get_stored_time();
	modemtime=NULL;
	while (modemtime==NULL) sendstring("AT+MCC?",&modemtime,NULL);
	day=(modemtime[0]-'0')*100+(modemtime[1]-'0')*10+modemtime[2]-'0';
	hour=(modemtime[4]-'0')*100+(modemtime[5]-'0')*10+modemtime[6]-'0';
	min=(modemtime[8]-'0')*100+(modemtime[9]-'0')*10+modemtime[10]-'0';
	sec=(modemtime[12]-'0')*100+(modemtime[13]-'0')*10+modemtime[14]-'0';
	free(modemtime);
	timedrift=time(NULL)-stored_time-day*3600*24-hour*3600-min*60-sec-1;
	if ( (day==255) && (hour==255) && (min==255) && (sec==255)) {
		logit(L_ERROR,"The modemclock has not been started.\n");
	} else {
		if ( ( timedrift > 3600 ) || ( timedrift < -3600 ) ) {
			logit(L_ERROR,"The modemclock has drifted more than an hour.\n");
		} else {
			logit(L_INFO,"The modemclock has drifted %d s.\n",(int)timedrift);
		}
	}
	return(0);
}


int setiospeed(speed_t baud) 
{
	if (cfsetispeed(&iosettings,baud)==-1) {
		logit(L_ERROR,"Error setting port ispeed.\n");
		exit(1);
	} else logit(L_DEBUG,"Port ispeed set to %d.\n",baud);
	if (cfsetospeed(&iosettings,baud)==-1) {
		logit(L_ERROR,"Error setting port ospeed.\n");
		exit(1);
	} else logit(L_DEBUG,"Port ospeed set to %d.\n",baud);
	
	tcflush(fd, TCIOFLUSH);
	tcsetattr(fd, TCSANOW, &iosettings);
	return(0);
}

int get_page(int n) {
	char shortstr[32];

	if (pages[n].filled) {
		logit(L_DEBUG,"Page %d already loaded.\n",n);
		page = pages[n].data;
		c_page = n;
		return(0);
	}
	logit(L_INFO,"Getting page %d ... ",n);
	sprintf(shortstr,"AT+MTP=%d",n);
	page = NULL;
	while (page == NULL) {
		page = getblock(shortstr,32776);
	}
	logit(L_INFO,"done!\n");
	c_page = n;
	page_size = getblock_return;
	/* compute real length */
	/* the return buffer looks as follows:
	   byte 0            checksum upper nibble
	   byte 1            checksum lower nibble
	   32768 bytes       data
	   \r\nOK\r\n
	   
	   together 32776 bytes
	*/
	pages[n].length = getblock_return-2-6;
	pages[n].data = malloc(pages[n].length*sizeof(char));
	bcopy(&(page[2]), pages[n].data, pages[n].length);
	pages[n].filled = 1;
	page = pages[n].data;
	logit(L_DEBUG,"get_page: page_nr=%d, page_length=%d\n",n,pages[n].length);
	return(0);
}

